#include <stdio.h>
#include <stm32f4xx.h>
#include <stm32f4xx_rcc.h>
#include <stm32f4xx_gpio.h>
#include <stm32f4xx_tim.h>

#include "misc.h"
#include "mpconfig.h"
#include "qstr.h"
#include "obj.h"
#include "encoder.h"
#include "led.h"

// encoder pins are:
//	PC6, PC7 - left encoder
//	PB8, PB9 - right encoder

#define LEFT_ENCODER_PORT GPIOC
#define LEFT_ENCODER_A_PIN GPIO_Pin_6
#define LEFT_ENCODER_B_PIN GPIO_Pin_7

#define RIGHT_ENCODER_PORT GPIOB
#define RIGHT_ENCODER_A_PIN GPIO_Pin_9
#define RIGHT_ENCODER_B_PIN GPIO_Pin_8

#define LEFT_ENC_A_EXTI_PIN (EXTI_PinSource6)
#define LEFT_ENC_B_EXTI_PIN (EXTI_PinSource7)
#define LEFT_ENC_A_EXTI_LINE (EXTI_Line6)
#define LEFT_ENC_B_EXTI_LINE (EXTI_Line7)
#define LEFT_ENC_EXTI_PORT (EXTI_PortSourceGPIOC)

#define RIGHT_ENC_A_EXTI_PIN (EXTI_PinSource9)
#define RIGHT_ENC_B_EXTI_PIN (EXTI_PinSource8)
#define RIGHT_ENC_A_EXTI_LINE (EXTI_Line9)
#define RIGHT_ENC_B_EXTI_LINE (EXTI_Line8)
#define RIGHT_ENC_EXTI_PORT (EXTI_PortSourceGPIOB)

#define ENCODER_EXTI_IRQN (EXTI9_5_IRQn)
#define ENCODER_EXTI_EDGE (EXTI_Trigger_Rising_Falling)


// Set up the four external interrupt pins
void encoder_init(void) {
	leftEncoder.state = 0;
	leftEncoder.counter = 0;
	leftEncoder.port = LEFT_ENCODER_PORT;
	leftEncoder.pinA = LEFT_ENCODER_A_PIN;
	leftEncoder.pinB = LEFT_ENCODER_B_PIN;

	rightEncoder.state = 0;
	rightEncoder.counter = 0;
	rightEncoder.port = RIGHT_ENCODER_PORT;
	rightEncoder.pinA = RIGHT_ENCODER_A_PIN;
	rightEncoder.pinB = RIGHT_ENCODER_B_PIN;

    // make it an input with no pull
	{
		GPIO_InitTypeDef GPIO_InitStructure;
		GPIO_InitStructure.GPIO_Pin  = LEFT_ENCODER_A_PIN;
		GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
		GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
		GPIO_Init(LEFT_ENCODER_PORT, &GPIO_InitStructure);
	}
	{
		GPIO_InitTypeDef GPIO_InitStructure;
		GPIO_InitStructure.GPIO_Pin  = LEFT_ENCODER_B_PIN;
		GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
		GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
		GPIO_Init(LEFT_ENCODER_PORT, &GPIO_InitStructure);
	}
	{
		GPIO_InitTypeDef GPIO_InitStructure;
		GPIO_InitStructure.GPIO_Pin  = RIGHT_ENCODER_A_PIN;
		GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
		GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
		GPIO_Init(RIGHT_ENCODER_PORT, &GPIO_InitStructure);
	}
	{
		GPIO_InitTypeDef GPIO_InitStructure;
		GPIO_InitStructure.GPIO_Pin  = RIGHT_ENCODER_B_PIN;
		GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
		GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
		GPIO_Init(RIGHT_ENCODER_PORT, &GPIO_InitStructure);
	}

    // the rest does the EXTI interrupt
    /* Enable SYSCFG clock */
    RCC_APB2PeriphClockCmd(RCC_APB2Periph_SYSCFG, ENABLE);

    /* Connect EXTI Line to GPIO pin */
    SYSCFG_EXTILineConfig(LEFT_ENC_EXTI_PORT, LEFT_ENC_A_EXTI_PIN);
    SYSCFG_EXTILineConfig(LEFT_ENC_EXTI_PORT, LEFT_ENC_B_EXTI_PIN);
    SYSCFG_EXTILineConfig(RIGHT_ENC_EXTI_PORT, RIGHT_ENC_A_EXTI_PIN);
    SYSCFG_EXTILineConfig(RIGHT_ENC_EXTI_PORT, RIGHT_ENC_B_EXTI_PIN);

	/* Configure EXTI Line */
	EXTI_InitTypeDef EXTI_InitStructure;
	EXTI_InitStructure.EXTI_Line = LEFT_ENC_A_EXTI_LINE | LEFT_ENC_B_EXTI_LINE | RIGHT_ENC_A_EXTI_LINE | RIGHT_ENC_B_EXTI_LINE;
	EXTI_InitStructure.EXTI_Mode = EXTI_Mode_Interrupt;
	EXTI_InitStructure.EXTI_Trigger = ENCODER_EXTI_EDGE;
	EXTI_InitStructure.EXTI_LineCmd = ENABLE;
	EXTI_Init(&EXTI_InitStructure);

    /* Enable and set EXTI15_10 Interrupt to the lowest priority */
    NVIC_InitTypeDef NVIC_InitStructure;
    NVIC_InitStructure.NVIC_IRQChannel = ENCODER_EXTI_IRQN;
    NVIC_InitStructure.NVIC_IRQChannelPreemptionPriority = 0x0F;
    NVIC_InitStructure.NVIC_IRQChannelSubPriority = 0x0F;
    NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
    NVIC_Init(&NVIC_InitStructure);
}

void encoder_edge_interrupt(encoder_t *encoder) {
	uint8_t s = encoder->state & 3;
	if (GPIO_ReadInputDataBit(encoder->port, encoder->pinA))
		s |= 4;
	if (GPIO_ReadInputDataBit(encoder->port, encoder->pinB))
		s |= 8;
	switch (s) {
		case 0: case 5: case 10: case 15:
			break;
		case 1: case 7: case 8: case 14:
			encoder->counter++; break;
		case 2: case 4: case 11: case 13:
			encoder->counter--; break;
		case 3: case 12:
			encoder->counter += 2; break;
		default:
			encoder->counter -= 2; break;
	}
	encoder->state = (s >> 2);
}

/******************************************************************************/
/* Micro Python bindings                                                      */

typedef struct _pyb_encoder_obj_t {
    mp_obj_base_t base;
    uint encoder_id;
	encoder_t *encoder;
} pyb_encoder_obj_t;

static void encoder_obj_print(void (*print)(void *env, const char *fmt, ...), void *env, mp_obj_t self_in, mp_print_kind_t kind) {
    pyb_encoder_obj_t *self = self_in;
    print(env, "<Encoder %lu: %ld>", self->encoder_id, self->encoder->counter);
}

static mp_obj_t encoder_obj_count(mp_obj_t self_in) {
    pyb_encoder_obj_t *self = self_in;
    return MP_OBJ_NEW_SMALL_INT(self->encoder->counter);
}

static MP_DEFINE_CONST_FUN_OBJ_1(encoder_obj_count_obj, encoder_obj_count);

static const mp_method_t encoder_methods[] = {
    { "count", &encoder_obj_count_obj },
    { NULL, NULL },
};

static const mp_obj_type_t encoder_obj_type = {
    { &mp_const_type },
    "Encoder",
    .print = encoder_obj_print,
    .methods = encoder_methods,
};

mp_obj_t pyb_Encoder(mp_obj_t encoder_id) {
    pyb_encoder_obj_t *o = m_new_obj(pyb_encoder_obj_t);
    o->base.type = &encoder_obj_type;
    o->encoder_id = mp_obj_get_int(encoder_id);
    switch ( o->encoder_id) {
        case 1: o->encoder = &leftEncoder; break;
        case 2: o->encoder = &rightEncoder; break;
    }
    return o;
}
